import atexit
from collections import defaultdict
import os
import time

DEBUG = os.getenv("DEBUG", None) is not None
DEBUG_COUNTS = defaultdict(int)
DEBUG_TIMES = defaultdict(float)


def print_debug_exit():
    if DEBUG:
        print(f"{'NAME':>50} : {'N_OCCUR':>8} {'TIME SPENT':>12} ms")
        for name, _ in sorted(DEBUG_TIMES.items(), key=lambda x: -x[1]):
            print(f"{name:>50} : {DEBUG_COUNTS[name]:>8} {DEBUG_TIMES[name]:>12.2f} ms")


atexit.register(print_debug_exit)


class Profile:
    def __init__(self, name):
        self.name = name

    def __enter__(self):
        if DEBUG:
            self.st = time.time()

    def __exit__(self, *junk):
        if DEBUG:
            et = (time.time() - self.st) * 1000

            DEBUG_COUNTS[self.name] += 1
            DEBUG_TIMES[self.name] += et



from dataclasses import dataclass
from math import ceil
from typing import List

from potator.extractors import EntitiesExtractor
from potator.indexer import Indexer, get_tokens_bounds
from potator.models import CloneData, DetectionResult, EntityData
from potator.similarity_metrics import jaccard
from potator.utils import sort_tokens_gtc, LANGUAGE_ORDER
from potator.profiler import Profile


def _validate_entity_candidate(entity: EntityData, candidate: EntityData) -> bool:
    return hash(entity) != hash(candidate) and entity.object_data.lang == candidate.object_data.lang and \
           entity.object_data.object_type == candidate.object_data.object_type


class Detector:
    def detect(self, directory: str, threshold: float, granularity: str) -> DetectionResult:
        raise NotImplemented


class NaiveDetector(Detector):
    def detect(self, directory: str, threshold: float, granularity: str) -> DetectionResult:
        with Profile("EntitiesExtractor.extract_data_from_directory"):
            files, files_data, entities = EntitiesExtractor.extract_data_from_directory(directory, granularity)

        clones = []
        with Profile("Search for clones in candidates set"):
            for i in range(len(entities)):
                entity = entities[i]
                for j in range(i + 1, len(entities)):
                    candidate = entities[j]
                    if not _validate_entity_candidate(entity, candidate):
                        continue

                    sim = jaccard(entity.bag_of_tokens, candidate.bag_of_tokens)
                    if sim > threshold:
                        clones.append(CloneData(entity, candidate, sim))

        clones = list(sorted(clones, reverse=True))

        return DetectionResult(clones, f"NaiveDetector(), directory: {directory}, threshold: {threshold}, "
                                       f"granularity: {granularity}")


@dataclass
class CandidateData:
    is_clone: bool
    score: float


def compute_candidate_similarity(tokens1: List[str], tokens2: List[str], ths: float, gtc: LANGUAGE_ORDER) \
        -> CandidateData:
    max_len = max(len(tokens1), len(tokens2))
    req_matches = ceil(max_len * ths)

    curr_matches = 0
    tok_pos1 = 0
    tok_pos2 = 0
    while tok_pos1 < len(tokens1) and tok_pos2 < len(tokens2):
        if min(len(tokens1) - tok_pos1, len(tokens2) - tok_pos2) + curr_matches >= req_matches:
            if tokens1[tok_pos1] == tokens2[tok_pos2]:
                curr_matches += 1
                tok_pos1 += 1
                tok_pos2 += 1
            else:
                if gtc[tokens1[tok_pos1]] < gtc[tokens2[tok_pos2]]:
                    tok_pos1 += 1
                else:
                    tok_pos2 += 1
        else:
            break

    if curr_matches >= req_matches:
        return CandidateData(True, curr_matches / max_len)
    else:
        return CandidateData(False, curr_matches / max_len)


class FilteringDetector(Detector):
    def __init__(self, max_l_depth: int = 1):
        self.max_l_depth = max_l_depth

    def detect(self, directory: str, threshold: float, granularity: str) -> DetectionResult:
        with Profile("EntitiesExtractor.extract_data_from_directory"):
            files, files_data, entities = EntitiesExtractor.extract_data_from_directory(directory, granularity)
        with Profile("sort_tokens_gtc"):
            gtc = sort_tokens_gtc(entities)
        with Profile("Indexer"):
            indexer = Indexer(entities, self.max_l_depth, threshold)

        clones = []
        considered_pairs = set()
        for i in range(len(entities)):
            entity = entities[i]
            lang = entity.object_data.lang
            tokens = entity.bag_of_tokens

            candidates = set()

            with Profile("Build candidates set"):
                for l_depth in range(1, self.max_l_depth + 1):
                    left_bound, right_bound = get_tokens_bounds(tokens, l_depth, threshold)
                    for token in tokens[left_bound: right_bound]:
                        candidates.update(indexer.get_entities_for_token(token, lang, l_depth))

            with Profile("Search for clones in candidates set"):
                for candidate in candidates:
                    if not _validate_entity_candidate(entity, candidate):
                        continue
                    if (entity, candidate) in considered_pairs or (candidate, entity) in considered_pairs:
                        continue
                    considered_pairs.add((entity, candidate))

                    candidate_data = compute_candidate_similarity(entity.bag_of_tokens, candidate.bag_of_tokens, threshold,
                                                                  gtc[entity.object_data.lang])
                    if candidate_data.is_clone:
                        clones.append(CloneData(entity, candidate, candidate_data.score))

        clones = list(sorted(clones, reverse=True))

        return DetectionResult(clones, f"FilteringDetector({self.max_l_depth}), directory: {directory}, "
                                       f"threshold: {threshold}, granularity: {granularity}")



from collections import defaultdict
from dataclasses import dataclass
from math import ceil
from typing import List, Dict, Tuple

from potator.models import EntityData


@dataclass
class Index:
    lang: str
    l_depth: int
    tokens2entities: Dict[str, List[EntityData]]


def get_tokens_bounds(tokens: List[str], l_depth: int, thr: float) -> Tuple[int, int]:
    left_bound = 0 if l_depth == 1 else len(tokens) - ceil(thr * len(tokens)) + l_depth - 1
    right_bound = len(tokens) - ceil(thr * len(tokens)) + l_depth
    return left_bound, right_bound


# TODO: maybe bounds should be on number of unique tokens. Think about that
def build_l_index(entities: List[EntityData], l_depth: int, thr: float = 0.8) -> Dict[str, Index]:
    tokens2entities = defaultdict(lambda: defaultdict(list))
    for entity in entities:
        lang = entity.object_data.lang
        tokens = entity.bag_of_tokens

        left_bound, right_bound = get_tokens_bounds(tokens, l_depth, thr)

        for token in set(tokens[left_bound: right_bound]):
            tokens2entities[lang][token].append(entity)

    return {lang: Index(lang, l_depth, tokens2entities[lang]) for lang in tokens2entities.keys()}


class Indexer:
    def __init__(self, entities: List[EntityData], max_l_depth: int = 1, threshold: float = 0.8):
        self.max_l_depth: int = max_l_depth
        self.threshold: float = threshold

        self._indexes: List[Dict[str, Index]] = []

        for l_depth in range(1, self.max_l_depth + 1):
            self._indexes.append(build_l_index(entities, l_depth, self.threshold))

    def get_entities_for_token(self, token, lang, depth):
        return self._indexes[depth - 1][lang].tokens2entities[token]



from argparse import ArgumentParser, Namespace

from potator.detectors import NaiveDetector, FilteringDetector
from potator.utils import write_formatted_detection_result, make_absolute_path
from potator.profiler import Profile


def parse_args() -> Namespace:
    parser = ArgumentParser(description="Find duplicate code in a directory")
    parser.add_argument("directory", type=str, help="directory with source code files")
    parser.add_argument("-d", "--detector", type=str, choices=["Naive", "Filtering"], help="type of detector to use")
    parser.add_argument("--depth", type=int, default=1, help="depth of adaptive prefix for FilteringDetector")
    parser.add_argument("-t", "--threshold", type=float, default=0.8, help="threshold of search")
    parser.add_argument("-g", "--granularity", type=str, default="functions", help="granularity of search")
    parser.add_argument("-o", "--out", type=str, default="output.html", help="dest of result of the search")
    args = parser.parse_args()
    return args


def main() -> None:
    program_args = parse_args()
    directory = make_absolute_path(program_args.directory)
    threshold = program_args.threshold
    granularity = program_args.granularity
    output = make_absolute_path(program_args.out)
    detector_type = program_args.detector
    depth = program_args.depth

    detector = NaiveDetector()
    if detector_type == "Filtering":
        detector = FilteringDetector(depth)

    with Profile("detection"):
        result = detector.detect(directory, threshold, granularity)

    print(f"Writing result to {output}...")
    write_formatted_detection_result(result, output)


if __name__ == "__main__":
    main()



from setuptools import setup
from setuptools.command.develop import develop
from setuptools.command.install import install
import pathlib

from potator.setup_tokenizer import setup_tokenizer


HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()


class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        develop.run(self)
        setup_tokenizer()


class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        setup_tokenizer()


setup(
    name='potator',
    description='',
    version='0.1.4',
    packages=['potator'],
    package_dir={'potator': 'potator'},
    install_requires=[
        'pygments',
        'joblib',
        'tqdm',
        'tree_sitter',
        'PyStemmer',
        'Cython'
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'potator = potator.__main__:main'
        ]
    },
    author="Yuriy Rogachev",
    long_description=README,
    long_description_content_type="text/markdown",
    license='MIT',
    url='https://github.com/otzhora/potator',
    download_url='https://github.com/otzhora/potator/releases/tag/v0.1.4',
    keywords=['STATIC-ANALYSIS', 'PLAGIARISM-DETECTION', 'PLAGIARISM-DETECTOR'],
    author_email='rogachev.yuiry28@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Quality Assurance',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'],
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand
    })
